/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/



#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include "ABKCommon/abkcommon.h"
#include "RBPlacement.h"

void RBPlacement::spaceCellsEqually(
			RBPSubRow& sr, 
			unsigned firstCellOffset, 
			unsigned numCells)
{
    abkassert(sr.getNumCells()-firstCellOffset >= numCells,
	"spaceCells called with numCells > #cells to right of startCell");	
    abkassert(numCells > 1, "spaceCells called for a single cell");

    std::vector<unsigned>::iterator cellsBegin =sr._cellIds.begin()+firstCellOffset;
    std::vector<unsigned>::iterator cellsEnd   =cellsBegin+numCells;

     double   rangeBegin,rangeEnd;
     rangeBegin  = _placement[*cellsBegin].x;	//left-edge of first cell
     unsigned numSpaces;

     if(cellsEnd == sr.cellIdsEnd())
     {
     	//right justify the last cell w/ the end of the subrow

	cellsEnd--;	//we will now only place n-1 cells...
	unsigned lastCell  = *(cellsEnd);
	double   lastCellW = _hgWDims->getNodeWidth(lastCell);
	_placement[lastCell].x = sr.getXEnd() -
				sr.getCoreRow().siteRoundUp(lastCellW);
	numCells--;
     }

     rangeEnd   = _placement[*cellsEnd].x;    //left-edge of next cell
     numSpaces  = numCells; // a space following each cell
		   		//including the right one
	
     double totalCellWidth = 0;
     std::vector<unsigned>::iterator srCell;
     for(srCell = cellsBegin; srCell != cellsEnd; srCell++)
     {
        double cellWidth = _hgWDims->getNodeWidth(*srCell);
        totalCellWidth += sr.getCoreRow().siteRoundUp(cellWidth);
     }

        //deltaSpace is the average space to put between each cell
     double deltaSpace = ((rangeEnd-rangeBegin)-totalCellWidth)/numSpaces;

     double leftEdge = rangeBegin;
     double cumExtraSpace = 0;  //extra space is accumulated as cells are
                                //placed. when the extra space is > k*sitewidth,
                                //k sites are skipped (for k any non-zero
                               //integer)

     for(srCell = cellsBegin; srCell != cellsEnd; srCell++)
     {
        unsigned cellId = *srCell;
        _placement[cellId].x = leftEdge;

        double cellWidth = _hgWDims->getNodeWidth(cellId);
        leftEdge += sr.getCoreRow().siteRoundUp(cellWidth);

        cumExtraSpace += deltaSpace;

        int numSitesOverlap = 
		static_cast<int>(floor(cumExtraSpace/sr.getSpacing()));
        leftEdge += sr.getSpacing() * numSitesOverlap;
        cumExtraSpace -= sr.getSpacing() * numSitesOverlap;
     }
}

void RBPlacement::spaceCellsWithPinDimensionsAlg1(
			RBPSubRow& sr, 
			unsigned firstCellOffset, 
			unsigned numCells)
{
    abkassert(sr.getNumCells()-firstCellOffset >= numCells,
	"spaceCells called with numCells > #cells to right of startCell");	
    abkassert(numCells > 1, "spaceCells called for a single cell");

    std::vector<unsigned>::iterator cellsBegin =sr._cellIds.begin()+firstCellOffset;
    std::vector<unsigned>::iterator cellsEnd   =cellsBegin+numCells;

    double   rangeBegin,rangeEnd;
    rangeBegin  = _placement[*cellsBegin].x;	//left-edge of first cell

    if(cellsEnd == sr.cellIdsEnd())
    {
         //right justify the last cell w/ the end of the subrow

        cellsEnd--;           //we will now only place n-1 cells...
        unsigned lastCell  = *(cellsEnd);
        double   lastCellW = _hgWDims->getNodeWidth(lastCell);
        _placement[lastCell].x = sr.getXEnd() -
                                 sr.getCoreRow().siteRoundUp(lastCellW);
        numCells--;
    }

    rangeEnd   = _placement[*cellsEnd].x +
                 sr.getCoreRow().siteRoundUp(
                     _hgWDims->getNodeWidth(*cellsEnd));
                                         //right-edge of next cell

    std::vector<unsigned>::iterator srCell;
    vector<double> xLeft(numCells + 1), xRight(numCells + 1); 
    vector<unsigned> leftPins(numCells + 1), rightPins(numCells + 1),
                     extremalPins(numCells + 1);
    vector<double> rightOffset(numCells + 1);
    vector<unsigned> cellCombined(numCells + 1, 1);
    double usedSpace = 0;

    unsigned cellOrder = 0, nextCellOrder = 0, cellId, totalExtremalPins = 0;
    for(srCell = cellsBegin; srCell != cellsEnd + 1; srCell++, cellOrder++)
    {
        cellId = (*srCell);
        const HGFNode& cell = _hgWDims->getNodeByIdx(cellId);

        itHGFEdgeLocal e;
        unsigned edgeOffset = 0;    // grab the extremal pin coordinates
        for(e = cell.edgesBegin(); e != cell.edgesEnd(); e++, edgeOffset++)
        {
            Point pinOffset =
                _hgWDims->edgesOnNodePinOffset(edgeOffset, cellId,
                                               _placement.getOrient(cellId));
            if(edgeOffset == 0)
                xLeft[cellOrder] = xRight[cellOrder] = pinOffset.x;
            else
            {
                xLeft[cellOrder]  = min(pinOffset.x, xLeft[cellOrder]);
                xRight[cellOrder] = max(pinOffset.x, xRight[cellOrder]);
            }
        }

        rightOffset[cellOrder] = 
            sr.getCoreRow().siteRoundUp(_hgWDims->getNodeWidth(cellId)) -
            xRight[cellOrder];
                         // leftOffset is the same as xLeft 

        if(cellOrder == 0) xLeft[cellOrder] = 0;
        if(cellOrder == numCells) 
        {
            xRight[cellOrder] += rightOffset[cellOrder];
            rightOffset[cellOrder] = 0;
        }

        edgeOffset = 0;            // calculate the number of extremal pins
        leftPins[cellOrder] = rightPins[cellOrder]
                            = extremalPins[cellOrder] = 0;
        for(e = cell.edgesBegin(); e != cell.edgesEnd(); e++, edgeOffset++)
        {
            Point pinOffset =
                _hgWDims->edgesOnNodePinOffset(edgeOffset, cellId,
                                               _placement.getOrient(cellId));
            if(pinOffset.x == xLeft[cellOrder] && cellOrder > 0)
                leftPins[cellOrder]++;
            if(pinOffset.x == xRight[cellOrder] && cellOrder < numCells)
                rightPins[cellOrder]++;
            if(pinOffset.x == xLeft[cellOrder]  && cellOrder > 0 || 
               pinOffset.x == xRight[cellOrder] && cellOrder < numCells)
                extremalPins[cellOrder]++;
        }
        usedSpace += xRight[cellOrder] - xLeft[cellOrder];
        totalExtremalPins += extremalPins[cellOrder];
    }

    bool needToRecompute = true;
    double deltaSpace = 0;

    while(needToRecompute)
    {
        needToRecompute = false;
        deltaSpace = ((rangeEnd - rangeBegin) - usedSpace)/
                        totalExtremalPins;

        for(cellOrder = 0; cellOrder < numCells; )
        {
            if (cellCombined[cellOrder] == 0) { cellOrder++; continue; }
            nextCellOrder = cellOrder + 1;
            while (nextCellOrder < numCells && cellCombined[nextCellOrder]==0)
                nextCellOrder ++;
            if (nextCellOrder >= numCells) break;
            if (rightOffset[cellOrder] + xLeft[nextCellOrder] < deltaSpace)
                { cellOrder++; continue; }

            needToRecompute = true;
            usedSpace += rightOffset[cellOrder];
            if (cellCombined[cellOrder] == 1)
                totalExtremalPins += leftPins[cellOrder] -
                                     extremalPins[cellOrder];
            else
                totalExtremalPins -= rightPins[cellOrder];
            if (cellCombined[nextCellOrder] == 1)
                totalExtremalPins += rightPins[nextCellOrder]-
                                     extremalPins[nextCellOrder];
            else
                totalExtremalPins -= leftPins[nextCellOrder];
            cellCombined[cellOrder] += cellCombined[nextCellOrder];
            cellCombined[nextCellOrder] = 0;
            xRight[cellOrder] += rightOffset[cellOrder]+xRight[nextCellOrder];
            rightOffset[cellOrder] = rightOffset[nextCellOrder];
            rightPins[cellOrder] = rightPins[nextCellOrder];
            extremalPins[cellOrder] = leftPins[cellOrder]+rightPins[cellOrder];
        }
    }

    double leftEdge = rangeBegin;
    double cumExtraSpace = 0;  //extra space is accumulated as cells are
                               //placed. when the extra space is > k*sitewidth,
                               //k sites are skipped (for k any non-zero
                               //integer)

    cellOrder = 0;
    double addedSpace = 0;

    for(srCell = cellsBegin; srCell != cellsEnd; srCell++, cellOrder++)
    {
        cellId = *srCell;
        if (cellCombined[cellOrder] > 0) leftEdge += addedSpace;
        _placement[cellId].x = leftEdge;

        double cellWidth = _hgWDims->getNodeWidth(cellId);
        leftEdge += sr.getCoreRow().siteRoundUp(cellWidth);

        if (cellCombined[cellOrder] == 0) continue;
        nextCellOrder = cellOrder + cellCombined[cellOrder];
        double interpinSpace =
            ((rightPins[cellOrder]    * extremalPins[cellOrder])     /
             (leftPins[cellOrder]     + rightPins[cellOrder])        +
             (leftPins[nextCellOrder] * extremalPins[nextCellOrder]) /
             (leftPins[nextCellOrder] + rightPins[nextCellOrder])) *
             deltaSpace;
        cumExtraSpace += interpinSpace - rightOffset[cellOrder]
                                       - xLeft[nextCellOrder];

        int numSitesOverlap = 
		static_cast<unsigned>(floor(cumExtraSpace/sr.getSpacing()));
        addedSpace = sr.getSpacing() * numSitesOverlap;
        cumExtraSpace -= addedSpace;
    }
}

void RBPlacement::spaceCellsWithPinDimensionsAlg2(
			RBPSubRow& sr, 
			unsigned firstCellOffset, 
			unsigned numCells)
{
    abkassert(sr.getNumCells()-firstCellOffset >= numCells,
	"spaceCells called with numCells > #cells to right of startCell");	
    abkassert(numCells > 1, "spaceCells called for a single cell");

    std::vector<unsigned>::iterator cellsBegin =sr._cellIds.begin()+firstCellOffset;
    std::vector<unsigned>::iterator cellsEnd   =cellsBegin+numCells;

    double   rangeBegin,rangeEnd;
    rangeBegin  = _placement[*cellsBegin].x;	//left-edge of first cell

    if(cellsEnd == sr.cellIdsEnd())
    {
     	//right justify the last cell w/ the end of the subrow

	cellsEnd--;	//we will now only place n-1 cells...
	unsigned lastCell  = *(cellsEnd);
	double   lastCellW = _hgWDims->getNodeWidth(lastCell);
	_placement[lastCell].x = sr.getXEnd() -
				sr.getCoreRow().siteRoundUp(lastCellW);
	numCells--;
    }

    rangeEnd   = _placement[*cellsEnd].x;    //left-edge of next cell

    double totalCellWidth = 0;
    vector<double> minFeasibleX(numCells), maxFeasibleX(numCells),
                   xLeft(numCells), xRight(numCells);
    vector<unsigned> numPins(numCells,0), numLeftPins(numCells);
    std::vector<unsigned>::iterator srCell;
    unsigned cellOrder = 0, cellId, totalPins = 0;
    for(srCell = cellsBegin; srCell != cellsEnd; srCell++, cellOrder++)
    {
        cellId = (*srCell);
        minFeasibleX[cellOrder] = totalCellWidth;
        double cellWidth = _hgWDims->getNodeWidth(cellId);
        totalCellWidth += sr.getCoreRow().siteRoundUp(cellWidth);
        const HGFNode& cell = _hgWDims->getNodeByIdx(cellId);

        itHGFEdgeLocal e;
        unsigned edgeOffset = 0;    // grab the extremal pin coordinates
        for(e = cell.edgesBegin(); e != cell.edgesEnd(); e++, edgeOffset++)
        {
            numPins[cellOrder]++;
            Point pinOffset =
                _hgWDims->edgesOnNodePinOffset(edgeOffset, cellId,
                                               _placement.getOrient(cellId));
            if(edgeOffset == 0)
                xLeft[cellOrder] = xRight[cellOrder] = pinOffset.x;
            else
            {
                xLeft[cellOrder]  = min(pinOffset.x, xLeft[cellOrder]);
                xRight[cellOrder] = max(pinOffset.x, xRight[cellOrder]);
            }
        }
        numLeftPins[cellOrder] = totalPins;
        totalPins += numPins[cellOrder];
    }

    double feasibleMax = rangeEnd - totalCellWidth;
    cellOrder = 0;
    for(srCell = cellsBegin; srCell != cellsEnd; srCell++, cellOrder++)
    {
        maxFeasibleX[cellOrder] = feasibleMax;
        double cellWidth = _hgWDims->getNodeWidth(*srCell);
        feasibleMax += sr.getCoreRow().siteRoundUp(cellWidth);
    }
        //deltaSpace is the average space to put between each pin
    double deltaSpace = (rangeEnd-rangeBegin)/(totalPins-1);

    double leftEdge = rangeBegin;
    cellOrder = 0;
    for(srCell = cellsBegin; srCell != cellsEnd; srCell++,cellOrder++)
    {
        cellId = *srCell;
        double leftPinCoord = numLeftPins[cellOrder] * deltaSpace,
               rightPinCoord = (numLeftPins[cellOrder] + numPins[cellOrder]
                                - 1) * deltaSpace;
        double idealX = (leftPinCoord  - xLeft[cellOrder] +
                         rightPinCoord - xRight[cellOrder]) / 2;

        int numSitesOverlap = 
		 	static_cast<unsigned>(floor(idealX/sr.getSpacing()));
        idealX = sr.getSpacing() * numSitesOverlap;   // snap to site

        if(idealX < minFeasibleX[cellOrder]) idealX = minFeasibleX[cellOrder];
        if(idealX > maxFeasibleX[cellOrder]) idealX = maxFeasibleX[cellOrder];
        if(idealX > leftEdge) leftEdge = idealX;
        _placement[cellId].x = leftEdge;

        double cellWidth = _hgWDims->getNodeWidth(cellId);
        leftEdge += sr.getCoreRow().siteRoundUp(cellWidth);
    }
}

