/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




// Created by Andrew Caldwell on 05/14/99
//  Derived from netStacks.h by Igor Markov

#ifndef _SPL_NETSTACKS_SROW_H_
#define _SPL_NETSTACKS_SROW_H_

#include "SmallPlacement/smallPlaceProb.h"
#include "SmallPlacement/xInterval.h"
#include "Ctainers/riskyStack.h"

class SmallPlacerNetStacks
{
  protected:

    struct EdgeAndOffset
    {
        unsigned edge;
        double   pinOffset;
    };
    
    unsigned    _numNets;
        
        // each element of _netData stores the xInterval for one hyperedge
        // at some point in the recursion.
    XInterval*  _netData;	//netData points to an array of XIntervals
				//of size #edges*#nodes
        
        // _netStacks[k] is a pointer to the element of _netData
        // describing edge k at the current moment in the recursion.
    XInterval** _netStacks;
    
	//_adjNets gives, for each node, the list of adjacent nets and
	//their offsets
    EdgeAndOffset*  _adjNets;
        
        // _firstNets[k] is a pointer to the first element of
        // _nets that refers to a net touching movable node k
    EdgeAndOffset** _firstNets;
        
    RiskyStack<double> _cost;

    double _leftEdge; //when addNodeToLeft is called, the cell being placed
		      //will have its origin at leftEdge, and leftEdge
		      //will increase by the cell's width
    double _rightEdge;	
    double _totalWidth;
    vector<double> _bloatWidths; //bloat cells
                                 //s.t. whitespace is equally distributed

    double _minWL; //min x-length any net can evaluate to


    void populate(const SmallPlacementProblem& problem);

  public:
        
    SmallPlacerNetStacks(const SmallPlacementProblem& problem, double minWL);

    virtual ~SmallPlacerNetStacks();

        // return value is new cost
	// the NetStacks maintain the left and right edges, at which
	// the cell will be placed
    double   addNodeToLeft (unsigned idx);	//idx of the node to add
    double   addNodeToRight(unsigned idx);

    double   removeNodeFromLeft (unsigned idx);
    double   removeNodeFromRight(unsigned idx);

    double   getCost() const { return _cost.back(); }
    const RiskyStack<double>& getCostStack() const { return _cost;}

    double   getLeftEdge()  const { return _leftEdge; }
    double   getRightEdge() const { return _rightEdge;}

    void     print();
};

inline double SmallPlacerNetStacks::addNodeToLeft (unsigned idx)
{
    //cout<<" AddNodeLeft|  node:"<<idx<<" at:"<<_leftEdge<<endl;

    double newCost = _cost.back();

    for(EdgeAndOffset *pt=_firstNets[idx]; pt!=_firstNets[idx+1]; pt++)
    {
	typedef XInterval* ivlPt;

		//subtract the previous edge cost
        ivlPt& netIvl = _netStacks[pt->edge];
	if(!netIvl->isEmpty())
	    newCost  -= max(netIvl->getLength(), _minWL);

		//compute the new edge cost, and add it 
	netIvl[1]  = netIvl[0];
        netIvl[1] += (_leftEdge + pt->pinOffset);
	newCost   += max(netIvl[1].getLength(), _minWL);
	
	netIvl++;
    }

    _cost.push_back(newCost);
    _leftEdge += _bloatWidths[idx];

    return newCost;
}

inline double SmallPlacerNetStacks::addNodeToRight(unsigned idx)
{
    _rightEdge -= _bloatWidths[idx];

    //cout<<" AddNodeRight|  node:"<<idx<<" at:"<<_rightEdge<<endl;

    double newCost = _cost.back();

    for(EdgeAndOffset *pt=_firstNets[idx]; pt!=_firstNets[idx+1]; pt++)
    {
        typedef XInterval* ivlPt;

                //subtract the previous edge cost
        ivlPt& netIvl = _netStacks[pt->edge];
        if(!netIvl->isEmpty())
            newCost  -= max(netIvl->getLength(), _minWL);

                //compute the new edge cost, and add it
        netIvl[1]  = netIvl[0];
        netIvl[1] += (_rightEdge + pt->pinOffset);
        newCost   += max(netIvl[1].getLength(), _minWL);

        netIvl++;
    }

    _cost.push_back(newCost);
    return newCost;
}

inline double SmallPlacerNetStacks::removeNodeFromLeft (unsigned idx)
{
    for(EdgeAndOffset *pt=_firstNets[idx]; pt!=_firstNets[idx+1]; pt++)
        --_netStacks[pt->edge];

    _leftEdge -= _bloatWidths[idx];

    _cost.pop_back();
    return _cost.back();
}

inline double SmallPlacerNetStacks::removeNodeFromRight(unsigned idx) 
{
    for(EdgeAndOffset *pt=_firstNets[idx]; pt!=_firstNets[idx+1]; pt++)
        --_netStacks[pt->edge];

    _rightEdge += _bloatWidths[idx];

    _cost.pop_back();
    return _cost.back();
}


#endif
